from typing import List, Optional

from pydantic import BaseModel

from .types import TaskCall


class CodeContextItem(BaseModel):
    content: str
    type: Optional[str] = None
    filepath: Optional[str] = None


class CodeContext(BaseModel):
    items: List[CodeContextItem]


class CodeTaskAttributes(BaseModel):
    prefix: str
    suffix: Optional[str] = None
    filepath: Optional[str] = None
    context: Optional[CodeContext] = None
    useControl: Optional[bool] = None
    provideGenerationDetails: Optional[bool] = None


# One-line


class CodeOneLineAllCompleteTask:
    """One-line code completion task for any language."""

    name = "code-one-line-all-complete"

    @classmethod
    def default(
        cls,
        prefix: str,
        suffix: Optional[str] = None,
        filepath: Optional[str] = None,
        context: Optional[CodeContext] = None,
        use_control: Optional[bool] = None,
        provide_generation_details: Optional[bool] = None,
    ) -> TaskCall:
        """
        Args:
            prefix: Code before cursor
            suffix: Code after cursor
            filepath: Relative path of the file, where completion is requested, with filename and extension
            context: Gathered context for code completion
            use_control: Whether to use control model or not
            provide_generation_details: Include logits from filter model in response
        """
        return TaskCall(
            id=f"{cls.name}:default",
            parameters=CodeTaskAttributes(
                prefix=prefix,
                suffix=suffix,
                filepath=filepath,
                context=context,
                useControl=use_control,
                provideGenerationDetails=provide_generation_details,
            ),
        )


class CodeOneLineKotlinCompleteTask(CodeOneLineAllCompleteTask):
    """One-line code completion task for Kotlin."""

    name = "code-one-line-kt-complete"


class CodeOneLinePythonJetCompleteTask(CodeOneLineAllCompleteTask):
    """One-line code completion task for Python."""

    name = "code-one-line-python-jet-complete"


# Multi-line


class CodeMultiLineAllCompleteTask:
    """Multi-line code completion task for all languages."""

    name = "code-multi-line-all-complete"

    @classmethod
    def fast(cls, prefix: str, suffix: str) -> TaskCall:
        """
        Args:
            prefix: Code before cursor
            suffix: Code after cursor
        """
        return TaskCall(
            id=f"{cls.name}:fast",
            parameters=dict(prefix=prefix, suffix=suffix),
        )

    @classmethod
    def slow(cls, prefix: str, suffix: str) -> TaskCall:
        """
        Args:
            prefix: Code before cursor
            suffix: Code after cursor
        """
        return TaskCall(
            id=f"{cls.name}:slow",
            parameters=dict(prefix=prefix, suffix=suffix),
        )


class CodeMultiLineJetKotlinCompleteTask(CodeOneLineAllCompleteTask):
    """Multi-line code completion task for Kotlin."""

    name = "code-multi-line-jet-kotlin-complete"


class CodeMultiLineJetPythonCompleteTask(CodeOneLineAllCompleteTask):
    """Multi-line code completion task for Python."""

    name = "code-multi-line-jet-python-complete"



import logging
import warnings
from typing import Any, AsyncIterator, Dict, Iterator, Optional

from .._api import BaseAPI, BaseAsyncAPI
from .types import (
    TaskCall,
    TaskCallResponse,
    TaskRoster,
    TaskStreamData,
    TaskStreamExecutionMetadata,
    TaskStreamQuotaMetaData,
    TaskStreamText,
)

logger = logging.getLogger(__name__)


class TasksAPI(BaseAPI):
    """TaskAPI client."""

    def execute_stream(self, call: TaskCall) -> Iterator[TaskStreamData]:
        """Executes supplied task call in TaskAPI.

        Args:
            call: task call to execute
        Returns:
            Stream of `TaskStreamData` objects in a form of iterator.
        """
        call_body = call.serialize()
        logger.debug("Executing request to TaskAPI with body: %r", call_body)
        for sse_event in self.client.stream(
            method="POST",
            path="task/stream/v2",
            json=call_body,
        ):
            if event := _convert_sse_event(sse_event):
                yield event

    def execute(self, call: TaskCall) -> TaskCallResponse:
        """Executes supplied task call in TaskAPI.

        Args:
            call: task call to execute
        Returns:
            `TaskCallResponse` object that aggregates the streamed response from TaskAPI.
        """
        content = []
        quota_metadata = None
        execution_metadata = []

        for task_stream_data in self.execute_stream(call):
            if isinstance(task_stream_data, TaskStreamText):
                content.append(task_stream_data.content)
            elif isinstance(task_stream_data, TaskStreamQuotaMetaData):
                quota_metadata = task_stream_data
            elif isinstance(task_stream_data, TaskStreamExecutionMetadata):
                execution_metadata.append(task_stream_data)

        return TaskCallResponse(
            content="".join(content),
            quota_metadata=quota_metadata,
            execution_metadata=execution_metadata,
        )

    def roster(self) -> TaskRoster:
        """Returns all available task ids."""
        resp = self.client.request(
            method="GET",
            path="task/roster",
        )
        return TaskRoster.parse_raw(resp.read())


class AsyncTasksAPI(BaseAsyncAPI):
    """TaskAPI async client."""

    async def execute_stream(self, call: TaskCall) -> AsyncIterator[TaskStreamData]:
        """Executes supplied task call in TaskAPI.

        Args:
            call: task call to execute
        Returns:
            Stream of `TaskStreamData` objects in a form of async iterator.
        """
        call_body = call.serialize()
        logger.debug("Executing request to TaskAPI with body: %r", call_body)
        async for sse_event in self.client.stream(
            method="POST",
            path="task/stream/v2",
            json=call_body,
        ):
            if event := _convert_sse_event(sse_event):
                yield event

    async def execute(self, call: TaskCall) -> TaskCallResponse:
        """Executes supplied task call in TaskAPI.

        Args:
            call: task call to execute
        Returns:
            `TaskCallResponse` object that aggregates the streamed response from TaskAPI.
        """
        content = []
        quota_metadata = None
        execution_metadata = []

        async for task_stream_data in self.execute_stream(call):
            if isinstance(task_stream_data, TaskStreamText):
                content.append(task_stream_data.content)
            elif isinstance(task_stream_data, TaskStreamQuotaMetaData):
                quota_metadata = task_stream_data
            elif isinstance(task_stream_data, TaskStreamExecutionMetadata):
                execution_metadata.append(task_stream_data)

        return TaskCallResponse(
            content="".join(content),
            quota_metadata=quota_metadata,
            execution_metadata=execution_metadata,
        )

    async def roster(self) -> TaskRoster:
        """Returns all available task ids."""
        resp = await self.client.request(
            method="GET",
            path="task/roster",
        )
        return TaskRoster.parse_raw(await resp.aread())


def _convert_sse_event(sse_event: Dict[str, Any]) -> Optional[TaskStreamData]:
    logger.debug("Received SSE from TaskAPI: %r", sse_event)

    event_type = sse_event.pop("type")

    if event_type == "Content":
        return TaskStreamText.parse_obj(sse_event)
    elif event_type == "QuotaMetadata":
        return TaskStreamQuotaMetaData.parse_obj(sse_event)
    elif event_type == "ExecutionMetadata":
        return TaskStreamExecutionMetadata.parse_obj(sse_event)

    warnings.warn(f"Can't convert unknown task stream event type {event_type!r}")
    return None



import json
import time
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel

from ..types import Credit, Quota

TASK_TYPE_MAP = {
    str: "text",
    bool: "bool",
    int: "int",
    float: "double",
    bytes: "bytes",
    dict: "json",
}


class Key(BaseModel):
    fqdn: str
    type: str


class Value(BaseModel):
    type: str
    value: Any
    modified: Optional[int] = None


class Attributes(BaseModel):
    data: List[Union[Key, Value]]


class TaskCall(BaseModel):
    id: str
    parameters: Union[Dict[str, Any], BaseModel] = {}

    @classmethod
    def from_obj(cls, obj: Union[Dict[str, Any], "TaskCall"]) -> "TaskCall":
        if isinstance(obj, TaskCall):
            return obj
        return cls(**obj)

    def serialize(self) -> Dict[str, Any]:
        data: List[Union[Key, Value]] = []
        now = int(time.time() * 1000)

        parameters = self.parameters
        if isinstance(parameters, BaseModel):
            parameters = parameters.dict()

        for key, value in parameters.items():
            if value is None:
                continue

            parame_type = TASK_TYPE_MAP.get(type(value))
            if parame_type is None:
                raise ValueError(f"Unsupported task parameter type {parame_type!r}")

            data.append(Key(fqdn=key, type=parame_type))
            if parame_type == "json":
                value = json.dumps(value)
            data.append(Value(modified=now, type=parame_type, value=value))

        return {
            "task": self.id,
            "parameters": Attributes(data=data).dict(),
        }


class TaskStreamText(BaseModel):
    content: str


class TaskStreamQuotaMetaData(BaseModel):
    updated: Quota
    spent: Credit


class TaskStreamExecutionMetadata(BaseModel):
    attributes: Attributes


TaskStreamData = Union[TaskStreamText, TaskStreamQuotaMetaData, TaskStreamExecutionMetadata]


class TaskCallResponse(BaseModel):
    content: str
    quota_metadata: Optional[TaskStreamQuotaMetaData] = None
    execution_metadata: List[TaskStreamExecutionMetadata] = []


class TaskRoster(BaseModel):
    ids: List[str]
